#ifndef GUI_H
#define GUI_H

#include "SDL2/SDL.h"
#include "point2D.h"
#include "console.h"
#include "math.h"
#include <iostream>
#include <vector>
#include <algorithm>



using namespace std;

class Color{
    public:
    Uint8 r,g,b,w;
    
    Color(const Uint8& R,const Uint8& G,const Uint8& B,const Uint8& W = 255):r(R),g(G),b(B),w(W){};

};


const Color RGB_Red(255,0,0);
const Color RGB_Green(0,255,0);
const Color RGB_Blue(0,0,255);
const Color RGB_Black(0,0,0);
const Color RGB_White(255,255,255);


class GUI{
    SDL_Window *win;
    SDL_Renderer *ren;
    SDL_Event e;
    bool open;
    
    public:    
    
    int w,h;
    
    float scroll;
    int x,y;
    bool left = false;
    bool right = false;
    
    GUI(char* name,int w,int h):win(NULL),ren(NULL),open(true),scroll(1),w(w),h(h),x(0),y(0){
        if(SDL_Init(SDL_INIT_EVERYTHING) < 0){
            cout << CONSOLE_RED << "SDL_Init error " << SDL_GetError() << CONSOLE_RESET << endl;
            fail();
        }
        
        win = SDL_CreateWindow(name, 100, 100, w, h, SDL_WINDOW_SHOWN);
        if (win == NULL){
            cout << CONSOLE_RED << "Error in creating window " << SDL_GetError() << CONSOLE_RESET << endl;
            fail();
        }
        
        ren = SDL_CreateRenderer(win, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
        if (ren == NULL){
            cout << CONSOLE_RED << "Error in creating renderer " << SDL_GetError() << CONSOLE_RESET << endl;
            fail();
        }
    }
    
    void loop(){
         while (SDL_PollEvent(&e)){
            switch(e.type){
                case SDL_QUIT:
                    close();
                    break;
                case SDL_MOUSEBUTTONDOWN:
                    if(e.button.button == SDL_BUTTON_LEFT)left = true;
                    if(e.button.button == SDL_BUTTON_RIGHT)right = true;
                    break;
                case SDL_MOUSEBUTTONUP:
                    if(e.button.button == SDL_BUTTON_LEFT)left = false;
                    if(e.button.button == SDL_BUTTON_RIGHT)right = false;
                    break;
                case SDL_MOUSEWHEEL:
                    scroll *= 1.0 - 0.03*e.wheel.y; 
                    break;
                case SDL_MOUSEMOTION:
                    x = e.motion.x;
                    y = e.motion.y;
                    break; 
            }
        }
    }
    
    void setColor(const Color& col){
        SDL_SetRenderDrawColor(ren,col.r,col.g,col.b,col.w);
    }
    
    void setPixel(int x,int y){
        SDL_RenderDrawPoint(ren,x,y);
    }
    
    void clear(const Color& col){
        setColor(col);
        SDL_RenderClear(ren);
    }
    
    void drawLine(int x1,int y1,int x2,int y2,const Color& col){
        setColor(col);
        SDL_RenderDrawLine(ren,x1,y1,x2,y2);
    }
    
    void fillCircle(int x,int y,int r,const Color& col){
        setColor(col);
        for(int i = -r;i < r;i++){
            for(int j = -r;j < r;j++){
                int d = i*i+j*j;
                if(d <= r*r)setPixel(x+i,y+j);
            }
        }
    }
    
    void drawLines(SDL_Point* points,int n,const Color& col){
        setColor(col);
        SDL_RenderDrawLines(ren,points,n);
    }
    
    void render(){
        SDL_RenderPresent(ren);
    }
    
    void close(){
        open = false;
        SDL_DestroyRenderer(ren);
        SDL_DestroyWindow(win);
    }
    
    bool isOpen(){
        return open;
    }

};



class Plot{
    float centerX;
    float centerY;
    float step;
    int sizeX;
    int sizeY;
    int Ox;
    int Oy;
    GUI g;
    Color background;
    Color axes;
    Color helpLines;
    
    
    struct func{
        double (*f)(double);
        Color color;
        SDL_Point* points;
        func(double (*F)(double),const Color& c,int w):f(F),color(c){
            points = (SDL_Point*)malloc(sizeof(SDL_Point)*w);
            for(int a = 0;a < w;a++)points[a].x = a;
        }
    };
    
    
    vector<func>functions;
    
    public:
    
    Plot(char* n,int x,int y):centerX(0),centerY(0),sizeX(x),sizeY(y),Ox(0),Oy(0),g(n,x,y),background(0,0,0),axes(255,255,255),helpLines(30,30,30){
        step = 0.01*g.scroll;
    }
    
    void drawAxes(){
        g.clear(background);
        int s = world2plot(1);
        int cX = world2plot(centerX)+sizeX/2;
        int cY = world2plot(centerY)+sizeY/2;
        
        int smallSize = 0.1/step;
        
        if(s > 0){
            //X small
            for(int a = cX;a < sizeX;a += s){
                g.drawLine(a, sizeY-1, a, 0, helpLines);
                g.drawLine(a, cY-smallSize, a, cY+smallSize, axes);
            }
            for(int a = cX;a > 0;a -= s){
                g.drawLine(a, sizeY-1, a, 0, helpLines);
                g.drawLine(a, cY-smallSize, a, cY+smallSize, axes);
            }
            //Y small        
            for(int a = cY;a < sizeY;a += s){
                g.drawLine(0,a,sizeX,a,helpLines);
                g.drawLine(cX-smallSize,a,cX+smallSize,a,axes);
            }
            for(int a = cY;a > 0;a -= s){
                g.drawLine(0,a,sizeX,a,helpLines);
                g.drawLine(cX-smallSize,a,cX+smallSize,a,axes);
            }
        }   
        //X     
        g.drawLine(0,cY,sizeX,cY,axes);
        //Y
        g.drawLine(cX, sizeY, cX, 0, axes);       
    }
    
    
    int world2plot(float a){
        return a/step+0.5;
    }
    
    float plot2world(int a){
        return a*step;
    }
    
    void drawFunction(const func& f){
            for(int a = 0;a < sizeX;a++){
                float y = f.f(plot2world(a-sizeX/2-world2plot(centerX)));
                f.points[a].y = sizeY/2-world2plot(y)+world2plot(centerY);
                //points.push_back(point(a,sizeY/2-world2plot(y)+world2plot(centerY)));
                //pset(world2plot(centerX)+a,sizeY/2-world2plot(y)+world2plot(centerY),RGB_Blue);
            }
            g.drawLines(f.points,sizeX,f.color);
            //for(int a = 1;a < points.size();a++)g.drawLine(points[a-1].x,points[a-1].y,points[a].x,points[a].y,f.color);
    }
    
    
    void addFunction(double (*f)(double),Color color){
        functions.push_back(func(f,color,sizeX));
    }
    
    
    bool draw(){
        g.loop();        
        drawAxes();
        for(int a = 0;a < functions.size();a++)drawFunction(functions[a]);
        g.render();
        
        //zoom
        step = 0.01*g.scroll;
        
        //drag
        if(g.left){
            if(Ox == 0 && Oy == 0){
                Ox = g.x;
                Oy = g.y;
            }else{
                centerX += plot2world(g.x-Ox);
                centerY += plot2world(g.y-Oy);
                Ox = g.x;
                Oy = g.y;
            }
        }else{
            Ox = 0;
            Oy = 0;
        }
        
        return g.isOpen();
    } 
    
    void loop(){
        bool ok = true;
        while(ok)ok = draw();
    }   
};

struct Agent{
    float x,y,angle;
    Agent(float X,float Y,float f):x(X),y(Y),angle(f){};
};

class Poligon{
	public:

	Point2D* points;
	Point2DInt* transformed;
	bool* mask;
	Transformation2D* t;

	int min_x,max_x,min_y,max_y;
	int size;
	Color color;

	Poligon(Point2D* p,int s,Color c = RGB_Red):points(p),size(s),color(c){
		transformed = new Point2DInt[s];
		mask = new bool[s];
		t = new Transformation2D();
	}

	void transform(float scale,float centerX,float centerY,int sizeX,int sizeY){
		min_x =  32767;
		max_x = -32767;
		min_y =  32767;
		max_y = -32767;
		for(int a = 0;a < size;a++){
			t->transform(&points[a],&transformed[a],scale,centerX,centerY,sizeX,sizeY);
			if(transformed[a].x < min_x)min_x = transformed[a].x;
			if(transformed[a].x > max_x)max_x = transformed[a].x;
			if(transformed[a].y < min_y)min_y = transformed[a].y;
			if(transformed[a].y > max_y)max_y = transformed[a].y;
		}
	}

	void resetMask(){
		for(int a = 0;a < size;a++){
			mask[a] = false;
		}
	}

};

class Space{
    float centerX;
    float centerY;
    float step;
    int sizeX;
    int sizeY;
    int Ox;
    int Oy;
    public:
    GUI g;
    Color background;

    vector<float*>points;
    vector<float*>red_points;

    vector<Poligon*>poligons;
    Agent robot;
    

    
    Space(char* n,int x,int y):centerX(0),centerY(0),sizeX(x),sizeY(y),Ox(0),Oy(0),g(n,x,y),background(0,0,0),robot(0,0,0){
        step = 0.01*g.scroll;
    }
    
    
    int world2plot(float a){
        return a/step+0.5;
    }
    
    float plot2world(int a){
        return a*step;
    }
    
    void drawPoint(float& x,float& y){
        int X = world2plot(x+centerX)+sizeX/2;
        int Y = sizeY/2-world2plot(y+centerY);
        if(X > 0 && X < g.w && Y > 0 && Y < g.h)g.setPixel(X,Y);
    }

    void drawPoligon(Poligon* p){
		p->transform(1/step,centerX,centerY,sizeX,sizeY);

    	for(int i = p->min_y;i <= p->max_y;i++){
    		vector<int>xes;
    		for(int a = 0;a < p->size;a++){
    			int b;
    			if(a == 0){
    				b = p->size-1;
    			}else{
    				b = a-1;
    			}
    			if((i <= p->transformed[a].y) ^ (i <= p->transformed[b].y)){//there is an intersection
    				int delta_y = p->transformed[a].y - p->transformed[b].y;
    				if(delta_y == 0){
    					xes.push_back(p->transformed[a].x);
						xes.push_back(p->transformed[b].x);
    				}else{
    					int delta_x = p->transformed[a].x - p->transformed[b].x;
    					float k = float(delta_x)/delta_y;
    					xes.push_back(p->transformed[a].x + k*(i - p->transformed[a].y));
    				}
    			}
    		}
    		sort(xes.begin(), xes.end());
    		if(xes.size() > 0)for(int a = 0;a < xes.size()-1;a+=2){
    			for(int j = xes[a];j < xes[a+1];j++){
    				g.setPixel(j,i);
    			}
    		}
    	}
    }
    
    void addPoint(float* p){
        points.push_back(p);
    }    

    #define KINECT_FOV 0.5 //half of kinect fov in radians

    void drawAgent(){
        int X = world2plot(robot.x+centerX)+sizeX/2;
        int Y = sizeY/2-world2plot(robot.y+centerY);
        int len = world2plot(3.0);
        g.drawLine(X,Y,X+len*cos(robot.angle+KINECT_FOV),Y+len*sin(robot.angle+KINECT_FOV),RGB_Red);
        g.drawLine(X,Y,X+len*cos(robot.angle),Y+len*sin(robot.angle),RGB_Blue);
        g.drawLine(X,Y,X+len*cos(robot.angle-KINECT_FOV),Y+len*sin(robot.angle-KINECT_FOV),RGB_Red);

    }

    void setAgent(float X,float Y,float a){
        robot.x = X;
        robot.y = Y;
        robot.angle = a-PI/2.0;
    }
    
    bool draw(){
        g.loop();
        g.clear(background);
        drawAgent();
    	g.setColor(RGB_White);
        for(int a = 0;a < poligons.size();a++)drawPoligon(poligons[a]);
        g.setColor(RGB_White);
        for(int a = 0;a < points.size();a++)drawPoint(points[a][0],points[a][1]);
        g.setColor(RGB_Red);
        for(int a = 0;a < red_points.size();a++)drawPoint(red_points[a][0],red_points[a][1]);
        g.render();
        
        //zoom
        step = 0.01*g.scroll;
        
        //drag
        if(g.left){
            if(Ox == 0 && Oy == 0){
                Ox = g.x;
                Oy = g.y;
            }else{
                centerX += plot2world(g.x-Ox);
                centerY -= plot2world(g.y-Oy);
                Ox = g.x;
                Oy = g.y;
            }
        }else{
            Ox = 0;
            Oy = 0;
        }
        
        return g.isOpen();
    } 
    
    void loop(){
        bool ok = true;
        while(ok)ok = draw();
    }   
};

#endif