#ifndef TREE_H
#define TREE_H

#include <vector>
#include <algorithm>

using namespace std;

template<class t,int dimensions = 2>class kdTree{
    struct node{
        t data;
        node* left;
        node* right;
        node(t d):data(d),left(NULL),right(NULL){}
    };
    
    class compare{
        public:
        int dim;
        compare(int d):dim(d){}
        bool operator()(t i,t j){
            return (*i)[dim] < (*j)[dim];
        }
    };
    
public:  
    node* root;
    
    node* best;
    double best_d;
    int size = 0;

    kdTree(vector<t>&d){
        size = d.size();
        root = setup(0,d.size(),0,d);
    }

    kdTree(char* name,t (*readData)(ifstream&)){
        ifstream f(name,ios::binary);
        root = this->read(f,readData);
    }
    
    void add(t d){
        Add(root,0,d);
    }    
    
    t nearestNeighbor(t d,double radius = 0){
        if(radius > 0)best_d = radius;
        else best_d = 1e300;
        best = NULL;
        NearestNeighbor(root,d,0);
        if(best == NULL)return NULL;
        return best->data;
    }


    void save(char* name){
        ofstream f(name,ios::binary);
        Save(root,f);
    }

    vector<t> vectorize(){
        vector<t>res;
        Vectorize(root,res);
        return res;
    }

    t minimize(int dim){
        best_d = 1e300;
        Minimize(root,dim,0);
        return best->data;
    }



    t maximize(int dim){
        best_d = -1e300;
        Maximize(root,dim,0);
        return best->data;
    }

private:

    node* read(ifstream& f,t (*readData)(ifstream&)){
        node* n = new node(readData(f));
        bool left,right;
        f.read((char*)&left,sizeof(bool));
        if(left)n->left = this->read(f,readData);
        f.read((char*)&right,sizeof(bool));
        if(right)n->right = this->read(f,readData);
        return n;
    }
    
    node* setup(int start,int end,int depth,vector<t>&d){
        if(start == end)return NULL;
        sort(d.begin()+start,d.begin()+end,compare(depth%dimensions));
        int median = (start+end)/2;
        node* n = new node(d[median]);
        n->left = setup(start,median,depth+1,d);
        n->right = setup(median+1,end,depth+1,d);
        return n;
    }

    void Add(node* n,int depth,t d){
        node** act = NULL;
        if((*n->data)[depth%dimensions] >= (*d)[depth%dimensions]){
            act = &n->left;
            if((*n->data)[depth%dimensions] == (*d)[depth%dimensions])if(Distance(n->data,d) == 0)return;
        }else{
            act = &n->right;
        }
        if(*act == NULL){
            *act = new node(d);
            size++;
        }
        else Add(*act,depth+1,d);
    }
    
    double Distance(t A,t B){
        double d = 0;
        for(int a = 0;a < dimensions;a++)d += pow((*A)[a]-(*B)[a],2);
        return d;
    }

    void NearestNeighbor(node* act,t d,int depth){
        if(act == NULL)return;
        //explore current node
        double dist = Distance(act->data,d);
        if(dist < best_d){
            best_d = dist;
            best = act;
        }
        //determine where the data would fall if inserted
        int dim = depth%dimensions;
        node* close = NULL;
        node* far = NULL;
        if((*d)[dim] < (* act->data)[dim]){//left
            close = act->left;
            far = act->right;
        }else{//right
            close = act->right;
            far = act->left;
        }
        //explore the closest side
        NearestNeighbor(close,d,depth+1);
        //explore the other side only if needed (best distance > distance from the dimension splitting hyperplane)
        if(best_d >= pow((*d)[dim]-(*act->data)[dim],2))NearestNeighbor(far,d,depth+1);
    }

    void Save(node* n,ofstream& f){
        n->data->save(f);
        bool left = (n->left != NULL);
        bool right = (n->right != NULL);
        f.write((char*)&left,sizeof(bool));
        if(left)Save(n->left,f);
        f.write((char*)&right,sizeof(bool));
        if(right)Save(n->right,f);        
    }

    void Vectorize(node* n,vector<t>&res){
        res.push_back(n->data);
        if(n->left != NULL)Vectorize(n->left,res);
        if(n->right != NULL)Vectorize(n->right,res);
    }

    void Minimize(node* n,int dim,int k){
        bool explore = false;
        if(k%dimensions == dim){//if current node is splitting the selected dimension
            if(n->left != NULL){//if there is a smaller child exploreit
                Minimize(n->left,dim,k+1);
            }else{
                explore = true;
            }
        }else{
            if(n->left != NULL)Minimize(n->left,dim,k+1);
            if(n->right != NULL)Minimize(n->right,dim,k+1);
            explore = true;
        }
        if( explore && (*n->data)[dim] < best_d ){
            best_d = (*n->data)[dim];
            best = n;
        }
    }

    void Maximize(node* n,int dim,int k){
        bool explore = false;
        if(k%dimensions == dim){//if current node is splitting the selected dimension
            if(n->right != NULL){//if there is a smaller child exploreit
                Maximize(n->right,dim,k+1);
            }else{
                explore = true;
            }
        }else{
            if(n->left != NULL)Maximize(n->left,dim,k+1);
            if(n->right != NULL)Maximize(n->right,dim,k+1);
            explore = true;
        }
        if( explore && (*n->data)[dim] > best_d ){
            best_d = (*n->data)[dim];
            best = n;
        }
    }
};

#endif